class FileManager {

      constructor() {
            this.rootPath = null;
            this.folderPath = null;
      }

      initFolder() {
            myApp.applog('externalDataDirectory '+cordova.file.externalDataDirectory);
            myApp.applog('externalRootDirectory '+cordova.file.externalRootDirectory);
            myApp.applog('documentsDirectory '+cordova.file.documentsDirectory);
            myApp.applog('applicationStorageDirectory '+cordova.file.applicationStorageDirectory);
            
            this.rootPath = cordova.file.externalRootDirectory;
            this.folderPath = 'Pictures/Images_compress/';
            this.createFolder(this.folderPath);
      }

      /**
       * Convert a base64 string in a Blob according to the data and contentType.
       * 
       * @param b64Data {String} Pure base64 string without contentType
       * @param contentType {String} the content type of the file i.e (image/jpeg - image/png - text/plain)
       * @param sliceSize {Int} SliceSize to process the byteCharacters
       * @see http://stackoverflow.com/questions/16245767/creating-a-blob-from-a-base64-string-in-javascript
       * @return Blob
       */
      b64toBlob(b64Data, contentType, sliceSize) {
            contentType = contentType || '';
            sliceSize = sliceSize || 512;

            var byteCharacters = atob(b64Data); // Buffer.from(b64Data, 'base64');
            var byteArrays = [];

            for (var offset = 0; offset < byteCharacters.length; offset += sliceSize) {
                  var slice = byteCharacters.slice(offset, offset + sliceSize);

                  var byteNumbers = new Array(slice.length);
                  for (var i = 0; i < slice.length; i++) {
                        byteNumbers[i] = slice.charCodeAt(i);
                  }

                  var byteArray = new Uint8Array(byteNumbers);

                  byteArrays.push(byteArray);
            }

            var blob = new Blob(byteArrays, {type: contentType});
            return blob;
      }

      /**
      * Create a Image file according to its database64 content only.
      * 
      * @param folderpath {String} The folder where the file will be created
      * @param filename {String} The name of the file that will be created
      * @param content {Base64 String} Important : The content can't contain the following string (data:image/png[or any other format];base64,). Only the base64 string is expected.
      */
      savebase64AsImageFile(folderpath,filename,content,contentType){

            if (folderpath=='') {
                  folderpath = this.rootPath+this.folderPath;
            }

            myApp.applog("savebase64AsImageFile "+folderpath+' '+filename+' '+content+' '+contentType);   

            // Convert the base64 string in a Blob
            let dataBlob = this.b64toBlob(content,contentType);
            myApp.applog("savebase64AsImageFile "+dataBlob);
            this.saveBlobToFile(folderpath, filename, dataBlob);
      }

      saveBlobToFile(folderpath, filename, blobdata) {
            myApp.applog("Starting to write the file :3");

            window.resolveLocalFileSystemURL(folderpath, function(dir) {
                  myApp.applog("Access to the directory granted succesfully");
                  dir.getFile(filename, { create: true, exclusive: false}, function(file) {
                        myApp.applog("File created succesfully.");
                        file.createWriter(
                              function(fileWriter) {
                                    myApp.applog("Writing content to file");                              
                                    fileWriter.write(blobdata);
                                    myApp.appsuccess('Compressed image successfully saved'); 
                                    myApp.appsuccess('Compressed image successfully saved as '+filename+' in folder '+myFileManager.folderPath+' (in path '+myFileManager.rootPath+')');
                                    myApp.applog("Writing content to file 3"); 
                              }, 
                              function(res = 'empty'){
                                    myApp.apperror('Unable to create file '+filename+' in path '+ folderpath);
                                    myApp.applog(JSON.stringify(res));
                              }
                        );
                  }, 
                  function(res = 'empty'){
                        myApp.apperror('Unable to create file '+filename+' in path '+ folderpath);
                        myApp.applog(JSON.stringify(res));
                  }     
                  );
            },
            function(res = 'empty'){
                  myApp.apperror('Unable to access directory '+ folderpath);
                  myApp.applog(JSON.stringify(res));
            }     
            );
      }

      createFolder(folderpath) {

            myApp.applog('create folder '+folderpath);

            // does the path already exist ?
            window.resolveLocalFileSystemURL(this.rootPath+folderpath, 
                  function(dir) {
                        // folder exist
                        myFileManager.folderIsReady(folderpath);
                  },
                  function(e) {
                        // create parent folder 
                        // remove last slash
                        if (folderpath.lastIndexOf('/')==folderpath.length-1) { folderpath = folderpath.substring(0, folderpath.length-1) }

                        // separate folder from path
                        let tFolder = folderpath.substring(folderpath.lastIndexOf('/')+1);
                        let tParentFolder = folderpath.substring(0, folderpath.lastIndexOf('/')+1);

                        // does the parent path already exist ?
                        window.resolveLocalFileSystemURL(myFileManager.rootPath+tParentFolder, 
                              function(dirEntry) {
                                    // parent folder exist = create child folder
                                    dirEntry.getDirectory(tFolder, { create: true, exclusive:false }, 
                                          function (subDirEntry) {
                                                myFileManager.folderIsReady(folderpath+'/');
                                          },
                                          function(e) {
                                                myApp.apperror('Could not create folder '+tFolder+' in '+tParentFolder);
                                                myApp.applog(JSON.stringify(e));

                                                // fallback to app folder when external root can't be accessed
                                                myFileManager.rootPath = cordova.file.externalDataDirectory;
                                                myFileManager.createFolder(folderpath);
                                          }
                                    );
                              },
                              function(e) {
                                    // create parent folder             
                                    myFileManager.createFolder(tParentFolder);
                              }
                        );

                  }
            );

      }

      folderIsReady(folderpath) {

            myApp.applog('folder '+folderpath+' is ready');

            if (folderpath!=this.folderPath) {
                  myFileManager.createFolder(this.folderPath);
            } else {
                  myApp.applog('main folder is ready');
            }

      }

      writeFile(folderpath, filename, dataObj=false) {
            resolveLocalFileSystemURL(folderpath, 
                  function(dirEntry) {
                        dirEntry.getFile(filename, {create: true, exclusive: false},
                              function(fileEntry) {
                                    fileEntry.createWriter(function (fileWriter) {

                                          fileWriter.onwriteend = function(e) {
                                                myApp.applog("Successful file write "+folderpath+filename);
                                                myApp.applog(JSON.stringify(e) );    
                                                myFileManager.readFile(folderpath+filename);
                                          };
                              
                                          fileWriter.onerror = function (e) {
                                                myApp.applog("Failed file write: "+ folderpath+filename );  
                                                myApp.applog(JSON.stringify(e) );                                  
                                          };
                              
                                          // If data object is not passed in,
                                          // create a new Blob instead.
                                          if (!dataObj) {
                                          dataObj = new Blob(['some file data'], { type: 'text/plain' });
                                          }
                              
                                          fileWriter.write(dataObj);
                                    });
                              }
                              ,
                              function(e) {
                                    myApp.applog('cannot access file '+folderpath+filename+' - '+JSON.stringify(e));
                              }
                        );
                  }
            );
      }

      readFile(filepath) {

            resolveLocalFileSystemURL(filepath, 
                  function(fileEntry) {

                        fileEntry.file(
                              function (file) {
                                    var reader = new FileReader();
                              
                                    reader.onloadend = function(res) {
                                          myApp.applog("Successful file read: " + filepath);
                                          myApp.applog('result - '+this.result);
                                          myApp.applog(JSON.stringify(res) );
                                    };
                              
                                    reader.readAsText(file);
                        
                              }, 
                              function(e) {
                                    myApp.applog('cannot read file '+filepath+' - '+JSON.stringify(e));
                              }
                        );
                  },
                  function(e) {
                        myApp.applog('cannot access file '+filepath+' - '+JSON.stringify(e));
                  }
            );
      }


}
