class ImageProcessor {
      
      constructor() {

            this.destWidth = null;
            this.destHeight = null; 
            this.resizeMethod = null; 
            this.rotation = null; 
            this.imageFileType = null; 
            this.imageQuality = null;

            this.validFileTypes = new Array('.jpg', '.jpeg', '.gif', '.png', '.webp');

            this.processingImages = 0;

            this.canvas = null;
            this.canvasContext = null;
            this.canvasTemp = null;

            this.initCanvas();

            if (typeof cordova !== 'undefined') { 

                  // Initialize the plugin
                  cordova.openwith.init(()=>{myApp.applog('init openwith success!');}, (error)=>{myApp.applog('init openwith failed: ' + error)});

                  // Define your file handler
                  cordova.openwith.addHandler((intent)=>{ this.processOpenWith(intent); });

            }
      }

      initCanvas() {

            this.canvas= document.createElement('canvas');
            this.canvas.width = 800;
            this.canvas.height = 600;
            //document.getElementById('canvas_container').appendChild(this.canvas);

            this.canvasContext = this.canvas.getContext('2d');
            this.canvasContext.imageSmoothingEnabled = true;
            this.canvasContext.imageSmoothingQuality = 'high';

            // Huawei fix : execute a first drawImage on canvas
            this.canvasContext.drawImage(document.getElementById('loading_img'), 0, 0, this.canvas.width, this.canvas.height);
      
            // step canvas
            this.canvasTemp = document.createElement('canvas');

      }

      validateType(filename, type) {

            let ext = filename.substring(filename.lastIndexOf ('.')).toLowerCase();                  
            myApp.applog(type+' '+ext);

            if (type.indexOf('image')<0 || this.validFileTypes.includes(ext)===false) {
                  myApp.applog('forbidden file type');
                        
                  let errorStr = 'The file { filename } of type { type } is forbidden. <br>Please use one of the supported file types: '+(this.validFileTypes.toString().replace(',', ', '));
                  errorStr = errorStr.replace('{ filename }', filename);
                  errorStr = errorStr.replace('{ type }', type);
                  myApp.apperror(errorStr);
      
                  return false;              
            }

            return true;
           
      }
     
      processForm(event) {

            /* INTERRUPT PROCESS */
            if (document.getElementById('process').classList.contains('processing')) { 
                  myApp.apperror('Please wait for end of current process');
                  return false; 
            }
            
            let files = document.getElementById('files').files;
            if (event.type==='drop') {
                  files = new Array();
                  // Use DataTransferItemList interface to access the file(s)
                  [...event.dataTransfer.items].forEach((item, i) => {
                        // If dropped items aren't files, reject them
                        if (item.kind === "file") {
                              const file = item.getAsFile();                              
                              console.log(`… file[${i}].name = ${file.name}`);
                              files.push(file);
                        }
                  });
            }

            /* START PROCESS */
            document.getElementById('process').classList.add('processing');

            if (typeof cordova !== 'undefined') {	
                  // Create folder (could have been deleted by user)	
                  myFileManager.initFolder();
            }
            
            this.destWidth = document.getElementById('input_width').value;
            this.destHeight = document.getElementById('input_height').value;
            this.resizeMethod = document.querySelector('input[name="resize"]:checked').value;
            this.rotation = parseInt(document.querySelector('select[name="rotation"]').value);
            this.imageFileType = document.querySelector('input[name="format"]:checked').value;
            this.imageQuality = parseInt(document.getElementById('input_quality_percent').value)/100;

            myApp.applog('processForm '+this.imageFileType+' '+this.imageQuality);
                              
            for (let f=0;f<files.length;f++) {
                  
                  let imageFile = files[f];
                                    
                  if (this.validateType(imageFile.name, imageFile.type)===false) {                                       
                        continue;                        
                  }
                  
                  // display on DOM
                  let num = this.createDivDOM(imageFile.type, imageFile.name, imageFile.size);

                  this.processingImages++;
                  
                  let fileReader = new FileReader();
                  fileReader.onload = function(event) { myImgProcessor.readLoadedFile(event.target.result, num); }
                  fileReader.readAsDataURL(imageFile);              
                  
            }

            if (this.processingImages===0) {
                  this.endProcess();
            }

            return false;
            
      }

      createDivDOM(type, name, size) {

            let divDOM = document.createElement('div');
            let num = document.querySelectorAll('#results div').length+1;
            divDOM.id = 'image_'+num;
            myApp.applog('createDivDOM id='+divDOM.id);
            divDOM.className = 'loading image_process';

            let new_format = (this.imageFileType.indexOf('png')>=0) ? 'PNG':'JPG';
            if (new_format==='JPG') { new_format+= ' '+(100*this.imageQuality)+'%' }
            
            divDOM.innerHTML = myTranslator.translate('Processing')+' '+new_format;
            divDOM.innerHTML += '<br>'+myTranslator.translate('Step')+' 1/3 : '+myTranslator.translate('Reading file')+' '+num;
            divDOM.setAttribute('data-type', type);
            divDOM.setAttribute('data-name', name);
            divDOM.setAttribute('data-size', size);
            divDOM.setAttribute('data-quality', this.imageQuality);
            divDOM.setAttribute('data-format', this.imageFileType);
            document.getElementById('results').appendChild(divDOM);

            return num;
      }

      processOpenWith(intent) {

            myApp.applog('intent received');
            
            document.getElementById('files').value = '';
            document.getElementById('clear').dispatchEvent(new MouseEvent('click'));
            document.getElementById('process').dispatchEvent(new MouseEvent('click'));

            myApp.applog('  action: ' + intent.action); // type of action requested by the user
            myApp.applog('  exit: ' + intent.exit); // if true, you should exit the app after processing
            myApp.applog('  items: ' + intent.items.length);//JSON.stringify(intent.items));
            if (intent.items.length > 0) {
                  myApp.applog('  items: ' + intent.items.length);//JSON.stringify(intent.items));
                  for (var i = 0; i < intent.items.length; i++) {
                        
                        let item = intent.items[i];
                        let filename = item.path.replace(/^.*\/(.*)$/gim, '$1');
                        myApp.applog('load intent item: '+i+' '+item.type+' '+filename);

                        if (this.validateType(filename, item.type)===false) {                                       
                              continue;                        
                        }

                        this.processingImages++;                        
                                                
                        let num = 0;
                        try {
                              num = this.createDivDOM(item.type, filename, 0);
                        } catch(e) {
                              myApp.applog('createDivDOM error '+JSON.stringify(e)+' '+e.message);
                        }
                       
                        // use timeout to start processing after div creation
                        setTimeout(()=>{
                              cordova.openwith.load(item, (data, item)=>{

                                    // data is a long base64 string with the content of the file
                                    //myApp.applog("the item " + JSON.stringify(item));
                                    myApp.applog("the item "+i+" weights " + data.length + " bytes");
                                    let divDOM = document.getElementById('image_'+num);
                                    divDOM.setAttribute('data-size', data.length);
                                    
                                    try {
                                          this.readLoadedFile('data:'+item.type+';base64, '+data, num);
                                    } catch(e) {
                                          myApp.applog('readLoadedFile error '+JSON.stringify(e)+' '+e.message);
                                    }
                                    // "exit" when done.
                                    // Note that there is no need to wait for the upload to finish,
                                    // the app can continue while in background.
                                    if (intent.exit) { cordova.openwith.exit(); }
                              });
                        }, 250);
                        
                  }
            }
            else {
                  if (intent.exit) { cordova.openwith.exit(); }
            }
            
      }

      readLoadedFile(imgData, num) {

            let divDOM = document.getElementById('image_'+num);
            myApp.applog('divDOM image_'+num + ' isNull:' + (divDOM===null));
            divDOM.innerHTML += '<br>'+myTranslator.translate('Step')+' 2/3 : '+myTranslator.translate('Loading image')+' '+num;

            let imageElement = document.createElement('img');
            imageElement.onload = function(event) { myImgProcessor.processLoadedImage(event, num); }
            imageElement.src = imgData;
            
      }

      processLoadedImage(event, num) {

            let imageElement = event.target;

            let divDOM = document.getElementById('image_'+num);
            divDOM.innerHTML += '<br>'+myTranslator.translate('Step')+' 3/3 : '+myTranslator.translate('Processing image')+' '+num;

            let srcWidth = imageElement.naturalWidth;
            let srcHeight = imageElement.naturalHeight;

            this.canvasTemp.width = srcWidth;
            this.canvasTemp.height = srcHeight;
            
            // use timeout to display loading message before processing image
            setTimeout(() => {
                  
                  let srcWidthPostRotation = srcWidth;
                  let srcHeightPostRotation = srcHeight;
                  
                  if (!isNaN(this.rotation)) {
                        if (Math.abs(this.rotation)==90) {
                              // invert image sizes
                              srcWidthPostRotation = srcHeight;
                              srcHeightPostRotation = srcWidth;
                        }
                  }
                                          
                  let canvasSize = this.autoSizeWithinSrcRatio(srcWidthPostRotation, srcHeightPostRotation, this.destWidth, this.destHeight);		
                  myApp.applog('canvasSize = '+canvasSize.toString());
                  this.canvas.width = canvasSize[0];
                  this.canvas.height = canvasSize[1];

                  // fill canvas = clear canvas
                  this.canvasContext.fillStyle = document.getElementById('input_color').value;
                  this.canvasContext.fillRect(0, 0, this.canvas.width, this.canvas.height);
                  
                  let canvasWidthPostRotation = canvasSize[0];
                  let canvasHeightPostRotation = canvasSize[1];
                  
                  if (!isNaN(this.rotation)) {
                        if (Math.abs(this.rotation)==90) {
                              // invert sizes
                              canvasWidthPostRotation = canvasSize[1];
                              canvasHeightPostRotation = canvasSize[0];
                        }
                  }
                  
                  let imgDestRect = this.applyResizingMethod(this.resizeMethod, srcWidth, srcHeight, canvasWidthPostRotation, canvasHeightPostRotation);
                  myApp.applog('imgDestRect = '+imgDestRect.toString());
                  
                  this.canvasContext.save();
                  
                  // this.rotation 
                  if (!isNaN(this.rotation)) {
                        
                        // move origin to canvas center
                        this.canvasContext.translate(canvasSize[0]/2, canvasSize[1]/2);

                        // rotate
                        this.canvasContext.rotate(this.rotation * Math.PI / 180);
                                    
                        // translate origin back
                        this.canvasContext.translate(-canvasWidthPostRotation/2, -canvasHeightPostRotation/2);

                  }		

                  // copy image on canvas
                  //this.canvasContext.drawImage(imageElement, imgDestRect[0], imgDestRect[1], imgDestRect[2], imgDestRect[3]);

                  // Hack for linear filtering -> resize by half steps
                  //let this.canvasTemp = document.createElement('canvas');		
                  let canvasTempCxt = this.canvasTemp.getContext('2d');
                  canvasTempCxt.drawImage(imageElement, 0, 0, this.canvasTemp.width, this.canvasTemp.height);
                  
                  myApp.applog('resize step width ', (this.canvasTemp.width>imgDestRect[2]*2));

                  while(this.canvasTemp.width>imgDestRect[2]*2) {

                        myApp.applog('resize step width '+this.canvasTemp.width);
                        
                        let canvasTemp2 = document.createElement('canvas'); 
                        canvasTemp2.width = this.canvasTemp.width/2;
                        canvasTemp2.height = this.canvasTemp.height/2;
                        let canvasTemp2Cxt = canvasTemp2.getContext('2d');
                        canvasTemp2Cxt.drawImage(this.canvasTemp, 0, 0, canvasTemp2.width, canvasTemp2.height);

                        this.canvasTemp = canvasTemp2;
                  }

                  this.canvasContext.drawImage(this.canvasTemp, imgDestRect[0], imgDestRect[1], imgDestRect[2], imgDestRect[3]);
                                          
                  this.canvasContext.restore();
                                          
                  let dataURL = this.canvas.toDataURL(this.imageFileType, this.imageQuality);
                                          
                  // display on DOM	
                  let trueDivDOM = document.getElementById(divDOM.id); // divDOM needs to be re-fetch (what's left is only a copy)	
                  divDOM = trueDivDOM;	
                  divDOM.innerHTML = '';
                  divDOM.className = 'image_process';
                  let imageDOM = document.createElement('img');
                  
                  if (canvasSize[1]>200 && canvasSize[0]/canvasSize[1] < 280/200) {
                        canvasSize = this.autoSizeWithinSrcRatio(canvasSize[0], canvasSize[1], 'auto', 200);
                  }				
                  
                  imageDOM.width = canvasSize[0];
                  imageDOM.height = canvasSize[1];
                  imageDOM.src = dataURL;				
                  divDOM.appendChild(imageDOM);

                  let former_size = Math.round(100*divDOM.getAttribute('data-size') /1024 /1024)/100;
                  let new_size = Math.round(100*window.atob(dataURL.split(",")[1]).length /1024 /1024)/100;
                  let size_gain = -Math.round(100*(former_size-new_size)/former_size);
                  let size_gain_result = 'positive';
                  if (size_gain>0) { 
                        size_gain = '+'+size_gain; 
                        size_gain_result = 'negative';
                  }
                  let former_format = (divDOM.getAttribute('data-type').indexOf('png')>=0) ? 'PNG':'JPG';
                  let new_format = (divDOM.getAttribute('data-format').indexOf('png')>=0) ? 'PNG':'JPG';
                  if (new_format==='JPG') { new_format+= ' '+(100*divDOM.getAttribute('data-quality'))+'%' }

                  let sizeDOM = document.createElement('DIV');
                  sizeDOM.className = 'process_result';
                  sizeDOM.innerHTML = myTranslator.translate('From')+' <b>'+former_size+'</b> MB '+former_format+' <br>'+myTranslator.translate('To')+' <b>'+new_size+'</b> MB (<b class="gain '+size_gain_result+'">'+size_gain+'</b>%) '+new_format;

                  divDOM.appendChild(sizeDOM);

                  // UPDATE extension
                  let newFileName = divDOM.getAttribute('data-name');
                  newFileName = newFileName.substring(0, newFileName.lastIndexOf('.'));
                  if (this.imageFileType=='image/png') {
                        newFileName += '.png';
                  } else if (this.imageFileType=='image/jpeg') {
                        newFileName += '.jpg';
                  }
                  
                  let divLinks = document.createElement('div');
                  divLinks.className = 'links';
                  divDOM.appendChild(divLinks);

                  let savelink = document.createElement('a');
                  savelink.innerHTML = myTranslator.translate('Save file');
                  savelink.href = dataURL;
                  savelink.setAttribute('data-type', 'save');
                  savelink.download = newFileName;
                  divLinks.appendChild(savelink);

                  if (typeof cordova !== 'undefined') {
                        let sharelink = document.createElement('a');
                        sharelink.innerHTML = myTranslator.translate('Share file');
                        sharelink.href = dataURL;
                        sharelink.className = 'share';
                        sharelink.setAttribute('data-type', 'share');
                        sharelink.download = newFileName;
                        divLinks.appendChild(sharelink);
                  }

                  this.processingImages--;
                  this.endProcess();

            }, 500);

      }

      endProcess() {
            if (this.processingImages===0) {
                  document.getElementById('process').classList.remove('processing');
            }
      }

      autoSizeWithinSrcRatio(srcWidth, srcHeight, destWidth, destHeight) {
            /* one or two dest_size is missing -> fill it following ratio */ 
            let ratio = srcWidth/srcHeight;
            
            if (isNaN(parseInt(destWidth)) && isNaN(parseInt(destHeight))) {
                  destWidth = srcWidth;
                  destHeight = srcHeight;
            } else {
                  if (isNaN(parseInt(destWidth))) {
                        destWidth = destHeight * ratio;
                  }
                  if (isNaN(parseInt(destHeight))) {
                        destHeight = destWidth / ratio;
                  }
            }		
            
            let canvasSize = new Array(Math.round(destWidth), Math.round(destHeight));
            
            return canvasSize;
      }

      applyResizingMethod(resizeMethod, srcWidth, srcHeight, destWidth, destHeight) {
            
            let destRatio = destWidth/destHeight;
            let srcRatio = srcWidth/srcHeight;
            let destX = 0;
            let destY = 0;
                  
            if (destRatio!=srcRatio) {
                  switch(resizeMethod) {
                        case 'crop':
                              
                              if (destRatio>srcRatio) {
                                    // DEST WIDTH BASED //
                                    let ratioHeight = destWidth/srcRatio;
                                    destY = (destHeight - ratioHeight)/2;
                                    destHeight = ratioHeight;
                                                                              
                              } else {
                                    // DEST HEIGHT BASED //
                                    
                                    let ratioWidth = destHeight*srcRatio;
                                    destX = (destWidth - ratioWidth)/2;
                                    destWidth = ratioWidth;
                              }
                                                      
                              break;
                        
                        case 'full':
                              if (destRatio>srcRatio) {
                                    // DEST HEIGHT BASED //
                                    
                                    let ratioWidth = destHeight*srcRatio;
                                    destX = (destWidth - ratioWidth)/2;
                                    destWidth = ratioWidth;
                                    
                              } else {
                                    // DEST WIDTH BASED //
                                    let ratioHeight = destWidth/srcRatio;
                                    destY = (destHeight - ratioHeight)/2;
                                    destHeight = ratioHeight;
                              }
                              break;
                              
                        case 'deform':
                        default:
                              // KEEP DEFAULT //
                              break;
                  }
            }
            
            let imgDestRect = new Array(destX, destY, destWidth, destHeight);
            
            return imgDestRect;
      }
}
